#include <stdlib.h>
#include <fcntl.h>
#include "libgraph.h"
#include <alloc.h>
#include <fail.h> 
#include <X11/keysym.h>
#include "libgraph.h"



Display * grdisplay;            /* The display connection */
int grscreen;                   /* The screen number */
int grscreen_height;            /* The screen height */
int grscreen_width;             /* The screen width */
Colormap grcolormap;            /* The color map */
canvas *grwindows;              /* The graphics window */
int nb_windows=0;               /* The number of windows */
int current_window;	       /* The number of windows */
int grwhite, grblack;           /* Black and white pixels */
unsigned long grcolor;          /* Current drawing color */
char ValidWindow;



Bool gr_initialized = False;


static int gr_error_handler(), gr_ioerror_handler();
value gr_clear_graph();

#define set_camlwin_env "CAMLWIN="

value gr_open_graph(void)	/* ML */
{
  char *display_name="";
  int ret;
  Window root,tmp;
  int wx,wy,ww,wh;
  unsigned int border_w;
  unsigned int depth;
  XSizeHints hints;
  char *env, *env_home;


  if (gr_initialized)
  {
    gr_clear_graph();
  } 
  else 
  {
    
  env = getenv(camlwin_env);
  if(env == NULL)
  {
    env_home = getenv("HOME");
    if(env_home == NULL) 
    {
      fprintf(stderr,"Can't get HOME variable, please set CAMLWIN\n");
      gr_fail("%s","gr_open : Can't get HOME variable");
      return Val_unit;
    }
    env=(char *)malloc(strlen(set_camlwin_env)+strlen(env_home)+1);
    strcpy(env,set_camlwin_env);
    strcat(env,env_home);
    ret=putenv(env);
#ifdef DEBUG
    fprintf(stderr,"open: putenv de CAMLWIN returned %d\n",ret);
#endif
  }

     ValidWindow=FALSE;
    grwindows=(canvas *)stat_alloc(sizeof(canvas));
    nb_windows=0;

       /* Open the display */
    grdisplay = XOpenDisplay(display_name);
    if (grdisplay == NULL)
      gr_fail("Cannot open display %s", XDisplayName(display_name));
    grscreen = DefaultScreen(grdisplay);
    grblack = BlackPixel(grdisplay, grscreen);
    grwhite = WhitePixel(grdisplay, grscreen);
    grcolormap = DefaultColormap(grdisplay, grscreen);

    /* Set up the error handlers */
    XSetErrorHandler(gr_error_handler);
    XSetIOErrorHandler(gr_ioerror_handler);
  
    /* Initial drawing color is black */
    grcolor = grblack;
    
    root=XDefaultRootWindow(grdisplay);
    XGetGeometry(grdisplay,root,&tmp,&wx,&wy,&grscreen_width,&grscreen_height,
                 &border_w,&depth);


    /* Parse the geometry specification */
    hints.x = 10;
    hints.height = 10;
    hints.y = 10;
    hints.width = 10;;
    hints.flags = PPosition | PSize;
    hints.win_gravity = 0;

    ret = XWMGeometry(grdisplay, grscreen, "", "", BORDER_WIDTH,
                      &hints, &wx, &wy, &ww, &wh, &hints.win_gravity);
    if (ret & (XValue | YValue)) {
      hints.x = wx; hints.y = wy; hints.flags |= USPosition;
    }
    if (ret & (WidthValue | HeightValue)) {
      hints.width = ww; hints.height = wh; hints.flags |= USSize;
    }

 
    /* Create the on-screen window */
    grwindows[0].win =
      XCreateSimpleWindow(grdisplay, DefaultRootWindow(grdisplay),
                          hints.x, hints.y, hints.width, hints.height,
                          BORDER_WIDTH, grblack, grwhite);
    XSetStandardProperties(grdisplay,  grwindows[0].win, WINDOW_NAME, ICON_NAME,
                           None, NULL, 0, &hints);
    grwindows[0].gc = XCreateGC(grdisplay,  grwindows[0].win, 0, NULL);
    XSetBackground(grdisplay,  grwindows[0].gc, grwhite);
    XSetForeground(grdisplay,  grwindows[0].gc, grcolor);
    XFlush(grdisplay);


    /* The global data structures are now correctly initialized.
       In particular, gr_sigio_handler can now handle events safely. */
    gr_initialized = True;
        
    /* Set the default font */
    ValidWindow=TRUE;
    grwindows[0].grfont=NULL;
    current_window=0;
    gr_set_font(Atom(0), Atom (0), Atom(0));
    ValidWindow=FALSE;

    /* Arrange for I/O on the connection to trigger the SIGIO signal */
    ret = fcntl(ConnectionNumber(grdisplay), F_GETFL, 0);
    fcntl(ConnectionNumber(grdisplay), F_SETOWN, getpid());
  
  }
  /* Reset the color cache */
  gr_init_color_cache();
  return Val_unit;    
}

value gr_close_graph()
{
  int i;

  if (gr_initialized) {
    gr_initialized = False;

    for(i=0;i<nb_windows;i++)
    {
      XFreeFont(grdisplay, grwindows[i].grfont); 
      grwindows[i].grfont = NULL;
      XFreeGC(grdisplay, grwindows[i].gc);
      XDestroyWindow(grdisplay, grwindows[i].win);
    }
    stat_free(grwindows);
    nb_windows=0;
    XCloseDisplay(grdisplay);
  }
  return Val_unit;
}

value gr_clear_graph()
{
  gr_check_open();
  XSetForeground(grdisplay, grwindows[current_window].gc, grwhite);
  XFillRectangle(grdisplay, grwindows[current_window].win, 
		 grwindows[current_window].gc,
                 0, 0, grwindows[current_window].w, 	
		 grwindows[current_window].h);
  XSetForeground(grdisplay, grwindows[current_window].gc, grcolor);

  XFlush(grdisplay);
  return Val_unit;
}

value gr_screen_width()
{
#ifdef DEBUG
    fprintf(stderr,"screen_width: %d\n",grscreen_width);
#endif
  return Val_int(grscreen_width);
}

value gr_screen_height()
{
#ifdef DEBUG
    fprintf(stderr,"screen_height: %d\n",grscreen_height);
#endif
  return Val_int(grscreen_height);
}

value gr_flush()
{
  XFlush(grdisplay); 
  return Val_unit;
}




/* The gr_sigio_handler is called via the signal machinery in the bytecode
   interpreter. The signal system ensures that this function will be
   called either between two bytecode instructions, or during a blocking
   primitive. In either case, not in the middle of an Xlib call.
   (There is no blocking primitives in this library, not even
   wait_next_event, for various reasons.) */

void gr_handle_simple_event();

value gr_sigio_handler()
{
  XEvent grevent;

#ifdef DEBUG_EVENT
       fprintf(stderr,"gr_sigio_handler\n");
#endif
  if (gr_initialized && nb_windows>0) {
    while (XCheckMaskEvent(grdisplay, -1 /*all events*/, &grevent))
      gr_handle_simple_event(&grevent);
  }
  return Val_unit;
}

void gr_handle_simple_event(e)
     XEvent * e;
{
  int i;

  switch (e->type) 
  {
   case Expose:
        gr_enqueue_events(e->xexpose.window,Redraw,0,0,0,0);
        break;

  case ConfigureNotify:
          for (i=0; i<nb_windows; i++)
          if(grwindows[i].win==e->xconfigure.window) break;

          if(i != nb_windows) 
             grwindows[i].h = e->xconfigure.height;
          break;

   case DestroyNotify:
        gr_enqueue_events(e->xdestroywindow.window,Close,0,0,0,0);
        break;

   case ButtonPress:
        gr_enqueue_events(e->xbutton.window,Button_down,
                          e->xbutton.x,e->xbutton.y,
                          e->type==ButtonPress,
                          0);
        break;

   case ButtonRelease:
        gr_enqueue_events(e->xbutton.window,Button_up,
                          e->xbutton.x,e->xbutton.y,
                          e->type==ButtonPress,
                          0);
        break;

   case MotionNotify:
        break;

   case KeyPress: 
       { 
        KeySym thekey;
        char keytxt[256];
        int nchars;
        char * p;

        nchars = XLookupString(&(e->xkey), keytxt, sizeof(keytxt), &thekey, 0);
        if(IsModifierKey(thekey)) break;
        else
          if(IsCursorKey(thekey) || IsFunctionKey(thekey) || 
             IsMiscFunctionKey(thekey))
               gr_enqueue_events(e->xkey.window,Keypressed,
                                 0, 0, 0, thekey);
          else  for (p = keytxt; nchars > 0; p++, nchars--) 
                 gr_enqueue_events(e->xkey.window,Keypressed,
                                   0, 0, 0, (int)*p);
       }
       break;

  case MappingNotify:
    XRefreshKeyboardMapping(&(e->xmapping));
    break;
  }
}

/* Processing of graphic errors */

void gr_fail(fmt, arg)
     char * fmt, * arg;
{
  char buffer[1024];
  sprintf(buffer, fmt, arg);
  fprintf(stderr,fmt,arg);
  raise_with_arg(GRAPHIC_FAILURE_EXN, copy_string(buffer));
}

void gr_check_open()
{
  if (!gr_initialized) gr_fail("graphic screen not opened", NULL);
}

static int gr_error_handler(display, error)
     Display * display;
     XErrorEvent * error;
{
  char errmsg[512];
  XGetErrorText(error->display, error->error_code, errmsg, sizeof(errmsg));
  gr_fail("Xlib error: %s", errmsg);
  return 0;
}

static int gr_ioerror_handler(display)
     Display * display;
{
  gr_fail("fatal I/O error", NULL);
  return 0;
}









